/*
 *   khexedit - Versatile hex editor
 *   Copyright (C) 1999  Espen Sand, espensa@online.no
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "dialogmargin.h"
#include "hextoolwidget.h"
#include "kapp.h"

#define MAX(x,y) ((x) > (y) ? (x): (y))


CHexToolWidget::CHexToolWidget( QWidget *parent, const char *name )
  : QFrame( parent, name )
{
  setFrameStyle( QFrame::Panel | QFrame::Raised );
  
  const char *text;
  mUtilBox = new QGridLayout( this, 6, 7, OUTER_SEP, INNER_SEP );
  if( mUtilBox == 0 ) { return; }

  mUtilBox->setColStretch( 6, 10 );

  const char *msg1[4] = 
  { 
    i18n("Signed 8 bit"), i18n("Unsigned 8 bit"), 
    i18n("Signed 16 bit"), i18n("Unsigned 16 bit") 
  };

  const char *msg2[4] = 
  { 
    i18n("Signed 32 bit"), i18n("Unsigned 32 bit"), 
    i18n("32 bit float"), i18n("64 bit float") 
  };
  
  const char *msg3[4] = 
  { 
    i18n("Hexadecimal"), i18n("Octal"), 
    i18n("Binary"), i18n("Text")
  };

  QLabel *label1[4];
  QLabel *label2[4];
  QLabel *label3[4];

  int max1 = 0;
  int max2 = 0;
  int max3 = 0;
  for( uint i=0; i<4; i++ )
  {
    label1[i] = new QLabel( i18n( msg1[i] ), this );
    max1 = MAX( max1, label1[i]->sizeHint().width() );
    label1[i]->setAlignment( AlignRight|AlignVCenter );
    mUtilBox->addWidget( label1[i], i+1, 0 );
  
    mText1[i] = new QLabel( "XXXXXXXX", this );
    mText1[i]->setMinimumSize( mText1[i]->sizeHint() );
    mText1[i]->setMaximumWidth( mText1[i]->sizeHint().width() );
    mUtilBox->addWidget( mText1[i], i+1, 1 );

    label2[i] = new QLabel( i18n( msg2[i] ), this );
    max2 = MAX( max2, label2[i]->sizeHint().width() );
    label2[i]->setAlignment( AlignRight|AlignVCenter );
    mUtilBox->addWidget( label2[i], i+1, 2 );
  
    mText2[i] = new QLabel( "XXXXXXXXXXXX", this );
    mText2[i]->setMinimumSize( mText2[i]->sizeHint() );
    mText2[i]->setMaximumWidth( mText2[i]->sizeHint().width() );
    mUtilBox->addWidget( mText2[i], i+1, 3 );

    label3[i] = new QLabel( i18n( msg3[i] ), this );
    max3 = MAX( max3, label3[i]->sizeHint().width() );
    label3[i]->setAlignment( AlignRight|AlignVCenter );
    mUtilBox->addWidget( label3[i], i+1, 4 );
  
    mText3[i] = new QLabel( "XXXXXXXX", this );
    mText3[i]->setMinimumSize( mText3[i]->sizeHint() );
    mText3[i]->setMaximumWidth( mText3[i]->sizeHint().width() );
    mUtilBox->addWidget( mText3[i], i+1, 5 );
  }

  text = i18n("Use little endian decoding");
  mCheckIntelFormat = new QCheckBox( text, this );
  mCheckIntelFormat->setMinimumSize( mCheckIntelFormat->sizeHint() );
  connect( mCheckIntelFormat, SIGNAL(clicked()),this,SLOT(intelFormat()) );
  mUtilBox->addMultiCellWidget( mCheckIntelFormat, 5, 5, 0, 2, AlignLeft );

  text = i18n("Display unsigned numbers as hexadecimal");
  mCheckHexadecimal = new QCheckBox( text, this );
  mCheckHexadecimal->setMinimumSize( mCheckHexadecimal->sizeHint() );
  connect( mCheckHexadecimal, SIGNAL(clicked()),this,SLOT(unsignedFormat()) );
  mUtilBox->addMultiCellWidget( mCheckHexadecimal, 5, 5, 3, 5, AlignLeft );

  QFontMetrics fm( mText1[0]->font() );
  for( uint i=0; i<4; i++ )
  {
    mText1[i]->setFixedHeight( fm.height()+2 );
    mText2[i]->setFixedHeight( fm.height()+2 );
    mText3[i]->setFixedHeight( fm.height()+2 );
    label1[i]->setFixedHeight( fm.height()+2 );
    label2[i]->setFixedHeight( fm.height()+2 );
    label3[i]->setFixedHeight( fm.height()+2 );
    label1[i]->setFixedWidth( max1 + 5 );
    label2[i]->setFixedWidth( max2 + 5 );
    label3[i]->setFixedWidth( max3 + 5 );

    mText1[i]->setLineWidth( 1 );
    mText2[i]->setLineWidth( 1 );
    mText3[i]->setLineWidth( 1 );
    mText1[i]->setFrameStyle( QFrame::Panel | QFrame::Sunken );
    mText2[i]->setFrameStyle( QFrame::Panel | QFrame::Sunken );
    mText3[i]->setFrameStyle( QFrame::Panel | QFrame::Sunken );
    mText1[i]->setBackgroundColor( kapp->windowColor );
    mText2[i]->setBackgroundColor( kapp->windowColor );
    mText3[i]->setBackgroundColor( kapp->windowColor );    
    mText1[i]->setAlignment( AlignRight|AlignVCenter );
    mText2[i]->setAlignment( AlignRight|AlignVCenter );
    mText3[i]->setAlignment( AlignRight|AlignVCenter );
    mText1[i]->setMargin( 3 );
    mText2[i]->setMargin( 3 );
    mText3[i]->setMargin( 3 );
    mText1[i]->setText( "" );
    mText2[i]->setText( "" );
    mText3[i]->setText( "" );
  }

  mUtilBox->activate();
  setGeometry( x(), y(), minimumSize().width(), minimumSize().height() );

  connect( kapp, SIGNAL( kdisplayPaletteChanged() ),
	   SLOT( paletteChanged() ) );
  connect( kapp, SIGNAL( kdisplayFontChanged() ),
	   SLOT( fontChanged() ) );

  mCursorState.valid = false;
  mViewHexCaps = true;

  show();
}


CHexToolWidget::~CHexToolWidget( void )
{
}


void CHexToolWidget::writeConfiguration( KConfig &config )
{
  config.setGroup("Conversion" );
  config.writeEntry("LittleEndian",  mCheckIntelFormat->isChecked() );
  config.writeEntry("UnsignedAsHex", mCheckHexadecimal->isChecked() );
}

void CHexToolWidget::readConfiguration( KConfig &config )
{
  config.setGroup("Conversion" );
  bool s1 = config.readBoolEntry( "LittleEndian", true );
  bool s2 = config.readBoolEntry( "UnsignedAsHex", false );
  mCheckIntelFormat->setChecked( s1 );
  mCheckHexadecimal->setChecked( s2 );
}




void CHexToolWidget::cursorChanged( SCursorState &state )
{
  if( state.valid == true )
  {
    char buf[40];
    if( mCheckIntelFormat->isChecked() == true )
    {
      //
      // Assume Little Endian. This is the case for the Intel architecture. 
      // The 'state.data' stores the first byte from file/memory in its 
      // first array element.
      //

      unsigned int i,j,k;
      unsigned long long d;
      float fv;
      double dv;

      i = state.data[5]<<8 | state.data[4];
      j = state.data[7]<<8 | state.data[6];
      d = i | j<<16;
      
      i = state.data[1]<<8 | state.data[0];
      j = state.data[3]<<8 | state.data[2];
      k = i | j<<16;
      d = d<<32 | k;

      memcpy( &fv, &k, 4 );
      memcpy( &dv, &d, 8 );

      if( mCheckHexadecimal->isChecked() == true )
      {
	if( mViewHexCaps == true )
	{
	  sprintf( buf, "0x%02X", state.data[0] );
	  mText1[1]->setText( buf );
	  sprintf( buf, "0x%04X", (unsigned short)i );
	  mText1[3]->setText( buf );
	  sprintf( buf, "0x%08X", k );
	  mText2[1]->setText( buf );
	}
	else
	{
	  sprintf( buf, "0x%02x", state.data[0] );
	  mText1[1]->setText( buf );
	  sprintf( buf, "0x%04x", (unsigned short)i );
	  mText1[3]->setText( buf );
	  sprintf( buf, "0x%08x", k );
	  mText2[1]->setText( buf );
	}

	sprintf( buf, "%d", (char)state.data[0] );
	mText1[0]->setText( buf );
	sprintf( buf, "%d", (short)i );
	mText1[2]->setText( buf );
	sprintf( buf, "%d", (int)k );
	mText2[0]->setText( buf );
	sprintf( buf, "%E", fv );
	mText2[2]->setText( buf );
	sprintf( buf, "%E", dv );
	mText2[3]->setText( buf );
      }
      else
      {
	sprintf( buf, "%d", (char)state.data[0] );
	mText1[0]->setText( buf );
	sprintf( buf, "%u", state.data[0] );
	mText1[1]->setText( buf );
	sprintf( buf, "%d", (short)i );
	mText1[2]->setText( buf );
	sprintf( buf, "%u", (unsigned short)i );
	mText1[3]->setText( buf );
	sprintf( buf, "%d", (int)k );
	mText2[0]->setText( buf );
	sprintf( buf, "%u", k );
	mText2[1]->setText( buf );
	sprintf( buf, "%E", fv );
	mText2[2]->setText( buf );
	sprintf( buf, "%E", dv );
	mText2[3]->setText( buf );
      }
    }
    else
    {
      //
      // Assume Big Endian. This is the case for SUN machines (amongst others)
      // The 'state.data' stores the first byte from file/memory in its 
      // first array element.
      //
      unsigned int i,j,k;
      unsigned long long d;
      float fv;
      double dv;

      i = state.data[5] | state.data[4]<<8;
      j = state.data[7] | state.data[6]<<8;
      d = i<<16 | j;
      
      i = state.data[1] | state.data[0]<<8;
      j = state.data[3] | state.data[2]<<8;
      k = i<<16 | j;
      d = d<<32 | k;

      memcpy( &fv, &k, 4 );
      memcpy( &dv, &d, 8 );
      
      if( mCheckHexadecimal->isChecked() == true )
      {
	if( mViewHexCaps == true )
	{
	  sprintf( buf, "0x%02X", state.data[0] );
	  mText1[1]->setText( buf );
	  sprintf( buf, "0x%04X", (unsigned short)i );
	  mText1[3]->setText( buf );
	  sprintf( buf, "0x%08X", k );
	  mText2[1]->setText( buf );
	}
	else
	{
	  sprintf( buf, "0x%02x", state.data[0] );
	  mText1[1]->setText( buf );
	  sprintf( buf, "0x%04x", (unsigned short)i );
	  mText1[3]->setText( buf );
	  sprintf( buf, "0x%08x", k );
	  mText2[1]->setText( buf );
	}

	sprintf( buf, "%d", (char)state.data[0] );
	mText1[0]->setText( buf );
	sprintf( buf, "%d", (short)i );
	mText1[2]->setText( buf );
	sprintf( buf, "%d", (int)k );
	mText2[0]->setText( buf );
	sprintf( buf, "%E", fv );
	mText2[2]->setText( buf );
	sprintf( buf, "%E", dv );
	mText2[3]->setText( buf );
      }
      else
      {
	sprintf( buf, "%d", (char)state.data[0] );
	mText1[0]->setText( buf );
	sprintf( buf, "%u", state.data[0] );
	mText1[1]->setText( buf );
	sprintf( buf, "%d", (short)i );
	mText1[2]->setText( buf );
	sprintf( buf, "%u", (unsigned short)i );
	mText1[3]->setText( buf );
	sprintf( buf, "%d", (int)k );
	mText2[0]->setText( buf );
	sprintf( buf, "%u", k );
	mText2[1]->setText( buf );
	sprintf( buf, "%E", fv );
	mText2[2]->setText( buf );
	sprintf( buf, "%E", dv );
	mText2[3]->setText( buf );
      }
    }

    unsigned char data = (unsigned char)state.data[0];
    sprintf( buf, "%02X", data );
    mText3[0]->setText( buf );
    sprintf( buf, "%03o", data );
    mText3[1]->setText( buf );
    for( int i = 0; i < 8; i++ )
    {
      buf[7-i] = (data&(1<<i)) ? '1' : '0';
    }
    buf[8] = 0;
    mText3[2]->setText( buf );
    sprintf( buf, "%c", state.charValid == false ? '.' : data );
    mText3[3]->setText( buf );
  }
  else
  {
    for( int i=0; i<4; i++)
    {
      mText1[i]->setText( "" );
      mText2[i]->setText( "" );
      mText3[i]->setText( "" );
    }
  }

  mCursorState = state;
}



void CHexToolWidget::paletteChanged( void )
{
  for( int i=0; i<4; i++ )
  {
    mText1[i]->setBackgroundColor( kapp->windowColor );
    mText2[i]->setBackgroundColor( kapp->windowColor );
    mText3[i]->setBackgroundColor( kapp->windowColor );  
  }
}


void CHexToolWidget::fontChanged( void )
{
  QFontMetrics fm( mText1[0]->font() );
  for( int i=0; i<4; i++ )
  {
    mText1[i]->setFixedHeight( fm.height()+2 );
    mText2[i]->setFixedHeight( fm.height()+2 );
    mText3[i]->setFixedHeight( fm.height()+2 );
  }
}


void CHexToolWidget::intelFormat( void )
{
  cursorChanged( mCursorState );
}


void CHexToolWidget::unsignedFormat( void )
{
  cursorChanged( mCursorState );
}


void CHexToolWidget::resizeEvent( QResizeEvent *e )
{
}





